(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "./Utils/Constants.js", "../Utils/Utils.js", "./Container.js", "../Utils/EventDispatcher.js", "../Utils/NumberUtils.js"], factory);
    }
})(function (require, exports) {
    "use strict";
    Object.defineProperty(exports, "__esModule", { value: true });
    exports.Engine = void 0;
    const Constants_js_1 = require("./Utils/Constants.js");
    const Utils_js_1 = require("../Utils/Utils.js");
    const Container_js_1 = require("./Container.js");
    const EventDispatcher_js_1 = require("../Utils/EventDispatcher.js");
    const NumberUtils_js_1 = require("../Utils/NumberUtils.js");
    function getItemsFromInitializer(container, map, initializers, force = false) {
        let res = map.get(container);
        if (!res || force) {
            res = [...initializers.values()].map((t) => t(container));
            map.set(container, res);
        }
        return res;
    }
    async function getDataFromUrl(data) {
        const url = (0, Utils_js_1.itemFromSingleOrMultiple)(data.url, data.index);
        if (!url) {
            return data.fallback;
        }
        const response = await fetch(url);
        if (response.ok) {
            return response.json();
        }
        (0, Utils_js_1.getLogger)().error(`${Constants_js_1.errorPrefix} ${response.status} while retrieving config file`);
        return data.fallback;
    }
    class Engine {
        constructor() {
            this._configs = new Map();
            this._domArray = [];
            this._eventDispatcher = new EventDispatcher_js_1.EventDispatcher();
            this._initialized = false;
            this.plugins = [];
            this._initializers = {
                interactors: new Map(),
                movers: new Map(),
                updaters: new Map(),
            };
            this.interactors = new Map();
            this.movers = new Map();
            this.updaters = new Map();
            this.presets = new Map();
            this.effectDrawers = new Map();
            this.shapeDrawers = new Map();
            this.pathGenerators = new Map();
        }
        get configs() {
            const res = {};
            for (const [name, config] of this._configs) {
                res[name] = config;
            }
            return res;
        }
        get version() {
            return "3.0.1";
        }
        addConfig(config) {
            const name = config.name ?? "default";
            this._configs.set(name, config);
            this._eventDispatcher.dispatchEvent("configAdded", { data: { name, config } });
        }
        async addEffect(effect, drawer, refresh = true) {
            (0, Utils_js_1.executeOnSingleOrMultiple)(effect, (type) => {
                !this.getEffectDrawer(type) && this.effectDrawers.set(type, drawer);
            });
            await this.refresh(refresh);
        }
        addEventListener(type, listener) {
            this._eventDispatcher.addEventListener(type, listener);
        }
        async addInteractor(name, interactorInitializer, refresh = true) {
            this._initializers.interactors.set(name, interactorInitializer);
            await this.refresh(refresh);
        }
        async addMover(name, moverInitializer, refresh = true) {
            this._initializers.movers.set(name, moverInitializer);
            await this.refresh(refresh);
        }
        async addParticleUpdater(name, updaterInitializer, refresh = true) {
            this._initializers.updaters.set(name, updaterInitializer);
            await this.refresh(refresh);
        }
        async addPathGenerator(name, generator, refresh = true) {
            !this.getPathGenerator(name) && this.pathGenerators.set(name, generator);
            await this.refresh(refresh);
        }
        async addPlugin(plugin, refresh = true) {
            !this.getPlugin(plugin.id) && this.plugins.push(plugin);
            await this.refresh(refresh);
        }
        async addPreset(preset, options, override = false, refresh = true) {
            (override || !this.getPreset(preset)) && this.presets.set(preset, options);
            await this.refresh(refresh);
        }
        async addShape(shape, drawer, refresh = true) {
            (0, Utils_js_1.executeOnSingleOrMultiple)(shape, (type) => {
                !this.getShapeDrawer(type) && this.shapeDrawers.set(type, drawer);
            });
            await this.refresh(refresh);
        }
        clearPlugins(container) {
            this.updaters.delete(container);
            this.movers.delete(container);
            this.interactors.delete(container);
        }
        dispatchEvent(type, args) {
            this._eventDispatcher.dispatchEvent(type, args);
        }
        dom() {
            return this._domArray;
        }
        domItem(index) {
            const dom = this.dom(), item = dom[index];
            if (!item || item.destroyed) {
                dom.splice(index, 1);
                return;
            }
            return item;
        }
        getAvailablePlugins(container) {
            const res = new Map();
            for (const plugin of this.plugins) {
                plugin.needsPlugin(container.actualOptions) && res.set(plugin.id, plugin.getPlugin(container));
            }
            return res;
        }
        getEffectDrawer(type) {
            return this.effectDrawers.get(type);
        }
        getInteractors(container, force = false) {
            return getItemsFromInitializer(container, this.interactors, this._initializers.interactors, force);
        }
        getMovers(container, force = false) {
            return getItemsFromInitializer(container, this.movers, this._initializers.movers, force);
        }
        getPathGenerator(type) {
            return this.pathGenerators.get(type);
        }
        getPlugin(plugin) {
            return this.plugins.find((t) => t.id === plugin);
        }
        getPreset(preset) {
            return this.presets.get(preset);
        }
        getShapeDrawer(type) {
            return this.shapeDrawers.get(type);
        }
        getSupportedEffects() {
            return this.effectDrawers.keys();
        }
        getSupportedShapes() {
            return this.shapeDrawers.keys();
        }
        getUpdaters(container, force = false) {
            return getItemsFromInitializer(container, this.updaters, this._initializers.updaters, force);
        }
        init() {
            if (this._initialized) {
                return;
            }
            this._initialized = true;
        }
        async load(params) {
            const id = params.id ?? `tsparticles${Math.floor((0, NumberUtils_js_1.getRandom)() * 10000)}`, { index, url } = params, options = url ? await getDataFromUrl({ fallback: params.options, url, index }) : params.options;
            let domContainer = params.element ?? document.getElementById(id);
            if (!domContainer) {
                domContainer = document.createElement("div");
                domContainer.id = id;
                document.body.append(domContainer);
            }
            const currentOptions = (0, Utils_js_1.itemFromSingleOrMultiple)(options, index), dom = this.dom(), oldIndex = dom.findIndex((v) => v.id.description === id);
            if (oldIndex >= 0) {
                const old = this.domItem(oldIndex);
                if (old && !old.destroyed) {
                    old.destroy();
                    dom.splice(oldIndex, 1);
                }
            }
            let canvasEl;
            if (domContainer.tagName.toLowerCase() === "canvas") {
                canvasEl = domContainer;
                canvasEl.dataset[Constants_js_1.generatedAttribute] = "false";
            }
            else {
                const existingCanvases = domContainer.getElementsByTagName("canvas");
                if (existingCanvases.length) {
                    canvasEl = existingCanvases[0];
                    canvasEl.dataset[Constants_js_1.generatedAttribute] = "false";
                }
                else {
                    canvasEl = document.createElement("canvas");
                    canvasEl.dataset[Constants_js_1.generatedAttribute] = "true";
                    domContainer.appendChild(canvasEl);
                }
            }
            if (!canvasEl.style.width) {
                canvasEl.style.width = "100%";
            }
            if (!canvasEl.style.height) {
                canvasEl.style.height = "100%";
            }
            const newItem = new Container_js_1.Container(this, id, currentOptions);
            if (oldIndex >= 0) {
                dom.splice(oldIndex, 0, newItem);
            }
            else {
                dom.push(newItem);
            }
            newItem.canvas.loadCanvas(canvasEl);
            await newItem.start();
            return newItem;
        }
        loadOptions(options, sourceOptions) {
            for (const plugin of this.plugins) {
                plugin.loadOptions(options, sourceOptions);
            }
        }
        loadParticlesOptions(container, options, ...sourceOptions) {
            const updaters = this.updaters.get(container);
            if (!updaters) {
                return;
            }
            for (const updater of updaters) {
                updater.loadOptions && updater.loadOptions(options, ...sourceOptions);
            }
        }
        async refresh(refresh = true) {
            if (!refresh) {
                return;
            }
            this.dom().forEach((t) => t.refresh());
        }
        removeEventListener(type, listener) {
            this._eventDispatcher.removeEventListener(type, listener);
        }
        setOnClickHandler(callback) {
            const dom = this.dom();
            if (!dom.length) {
                throw new Error(`${Constants_js_1.errorPrefix} can only set click handlers after calling tsParticles.load()`);
            }
            for (const domItem of dom) {
                domItem.addClickHandler(callback);
            }
        }
    }
    exports.Engine = Engine;
});
